from __future__ import annotations

import ast
import asyncio
import hashlib
import shutil
import textwrap
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import anyio
import anyio.to_process
import msgspec
import msgspec.json
from anyio import Path

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Iterator

    from .config import Config


def _hash_ast(source: str) -> str:
    return hashlib.sha1(
        "".join(_stringify_ast(ast.parse(source, type_comments=True))).encode("utf-8")
    ).hexdigest()


def _normalize(lineend: str, value: str) -> str:
    # To normalize, we strip any leading and trailing space from
    # each line...
    stripped = [i.strip() for i in value.splitlines()]
    normalized = lineend.join(stripped)
    # ...and remove any blank lines at the beginning and end of
    # the whole string
    return normalized.strip()


def _stringify_ast(node: ast.AST, depth: int = 0) -> Iterator[str]:
    # this was taken from black
    # https://github.com/psf/black/blob/ddfecf06c13dd86205c851e340124e325ed82c5c/src/black/parsing.py#L143-L216
    # and serves as a workaround for generating a hash of an AST, because ast.unparse
    # is not available until Python 3.9
    """Simple visitor generating strings to compare ASTs by content."""

    if (
        isinstance(node, ast.Constant)
        and isinstance(node.value, str)
        and node.kind == "u"
    ):
        # It's a quirk of history that we strip the u prefix over here. We used to
        # rewrite the AST nodes for Python version compatibility and we never copied
        # over the kind
        node.kind = None

    yield f"{'  ' * depth}{node.__class__.__name__}("

    for field in sorted(node._fields):
        # TypeIgnore has only one field 'lineno' which breaks this comparison
        if isinstance(node, ast.TypeIgnore):
            break

        try:
            value = getattr(node, field)
        except AttributeError:
            continue

        yield f"{'  ' * (depth + 1)}{field}="

        if isinstance(value, list):
            for item in value:
                # Ignore nested tuples within del statements, because we may insert
                # parentheses and they change the AST.
                if (
                    field == "targets"
                    and isinstance(node, ast.Delete)
                    and isinstance(item, ast.Tuple)
                ):
                    for elt in item.elts:
                        yield from _stringify_ast(elt, depth + 2)

                elif isinstance(item, ast.AST):
                    yield from _stringify_ast(item, depth + 2)

        elif isinstance(value, ast.AST):
            yield from _stringify_ast(value, depth + 2)

        else:
            if (
                isinstance(node, ast.Constant)
                and field == "value"
                and isinstance(value, str)
            ):
                # Constant strings may be indented across newlines, if they are
                # docstrings; fold spaces after newlines when comparing. Similarly,
                # trailing and leading space may be removed.
                normalized = _normalize("\n", value)
            elif field == "type_comment" and isinstance(value, str):
                # Trailing whitespace in type comments is removed.
                normalized = value.rstrip()
            else:
                normalized = value
            yield f"{'  ' * (depth + 2)}{normalized!r},  # {value.__class__.__name__}"

    yield f"{'  ' * depth})  # /{node.__class__.__name__}"


@dataclass
class TransformationResult:
    source: Path
    target: Path
    transformed: bool


class Cache:
    def __init__(self) -> None:
        self.cache_dir = Path(".unasyncd_cache")
        self.cache_content_dir = self.cache_dir / "content"

    async def _init_cache_dir(self) -> None:
        await self.cache_content_dir.mkdir(exist_ok=True, parents=True)
        cachedir_tag = self.cache_dir / "CACHEDIR.TAG"
        gitignore_file = self.cache_dir / ".gitignore"
        if not await cachedir_tag.exists():
            await cachedir_tag.write_text(
                textwrap.dedent(
                    """Signature: 8a477f597d28d172789f06886806bc55
                    # This file is a cache directory tag created by unasyncd.
                    # For information about cache directory tags, see:
                    #	http://www.brynosaurus.com/cachedir/
                    """
                )
            )
        if not await gitignore_file.exists():
            await gitignore_file.write_text("*\n")

    async def set(self, key: str, data: str) -> None:
        await self._init_cache_dir()
        cache_path = self.cache_content_dir / key
        await cache_path.write_text(data)

    async def get(self, key: str) -> str | None:
        if path := await self.get_path(key):
            return await path.read_text()
        return None

    async def get_path(self, key: str) -> Path | None:
        cache_path = self.cache_content_dir / key
        if await cache_path.exists():
            return cache_path

        return None


class FileMeta(msgspec.Struct):
    signature: str
    content_hash: str


class File(Path):
    """
    Wrapper around an ``anyio.Path`` that retrieves and caches metadata and
    contents.
    """

    def __init__(self, path: str | Path):
        super().__init__(path)
        self._content: str | None = None
        self._signature: str | None = None
        self._content_hash: str | None = None

    async def get_content(self) -> str:
        if self._content is None:
            self._content = await self.read_text()
        return self._content

    async def get_signature(self) -> str:
        if self._signature is None:
            stat = await self.stat()
            self._signature = f"{stat.st_mtime}:{stat.st_size}:{stat.st_mode}"
        return self._signature

    async def get_content_hash(self) -> str:
        if self._content_hash is None:
            self._content_hash = _hash_ast(await self.get_content())
        return self._content_hash

    async def to_meta(self) -> FileMeta:
        return FileMeta(
            signature=await self.get_signature(),
            content_hash=await self.get_content_hash(),
        )

    async def meta_equal(self, other: Any) -> bool:
        if isinstance(other, FileMeta):
            return bool(
                await self.get_signature() == other.signature
                or await self.get_content_hash() == other.content_hash
            )

        return False


class Env:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.cache = Cache()
        self._meta: dict[str, FileMeta] | None = None
        self._executor = ProcessPoolExecutor()
        self._config_key = self.config.key()

    async def _config_changed(self) -> bool:
        """
        Return a ``bool`` indicating if the current configuration has changed since the
        last run, by comparing its has with the hash stored in the ``__config_key_``
        cache file
        """
        if key := await self.cache.get("__config_key__"):
            return key != self._config_key
        return True

    async def get_meta(self, file: File) -> FileMeta | None:
        """Return the :class:`FileMeta` for ``file`` from the cache. If no meta has been
        recorded, return ``None``
        """
        if self._meta is None:
            if meta_raw := await self.cache.get("meta.json"):
                self._meta = msgspec.json.decode(meta_raw, type=dict[str, FileMeta])
            else:
                self._meta = {}

        return self._meta.get(str(file))

    async def set_meta(self, file: File) -> None:
        """Set the :class:`FileMeta` for ``file``"""
        if self._meta is None:
            self._meta = {}

        self._meta[str(file)] = await file.to_meta()

    async def _cache_key_for(self, file: File) -> str:
        """Return the cache key for ``file``"""
        return hashlib.sha1(
            (await file.get_content_hash() + self._config_key).encode()
        ).hexdigest()

    async def transform(self, file: File) -> str:
        """
        Return the transformed contents of ``file`` according to the current
        configuration
        """
        from .transformers import TreeTransformer

        transformer = TreeTransformer(
            exclude=self.config.exclude.get(str(file), set()),
            transform_docstrings=self.config.transform_docstrings,
            extra_name_replacements=self.config.extra_replacements.get(str(file), {}),
            infer_type_checking_imports=self.config.infer_type_checking_imports,
            ruff_fix=self.config.ruff_fix,
            ruff_format=self.config.ruff_format,
            file_name=str(file),
        )

        content = await file.get_content()
        loop = asyncio.get_running_loop()
        transformed = await loop.run_in_executor(self._executor, transformer, content)
        if self.config.add_editors_note:
            note = (
                "# Do not edit this file directly. It has been autogenerated from\n"
                f"# {file}\n"
            )
            transformed = note + transformed

        return transformed

    async def files_diverged(self, source: File, target: File) -> bool:
        """Return a bool indicating if ``source`` or ``target`` have diverged from
        their expected state.

        The files are considered equal if since the last run:

        - ``source_file`` stats have not changed
        - ``source_file`` is AST-equivalent to its previous version
        - ``target_file`` does exist

        If the ``force_regen`` option is used, always return ``True``.
        """
        if self.config.force_regen:
            return True

        if not await source.meta_equal(await self.get_meta(source)):
            return True

        if not await target.exists():
            return True

        return False

    async def _restore_from_cache(self, source: File, target: File) -> bool:
        """If a cached transformation for ``source`` exists, copy it to ``target``.
        If ``force_regen`` is ``True``, do nothing.
        """
        if self.config.force_regen:
            return False

        cached_file = await self.cache.get_path(await self._cache_key_for(source))
        if not cached_file:
            return False

        await target.parent.mkdir(parents=True, exist_ok=True)
        await anyio.to_thread.run_sync(shutil.copyfile, cached_file, target)
        return True

    async def unasync_file(
        self,
        source_path: str | Path,
        target_path: str | Path,
    ) -> TransformationResult:
        """
        Transform ``source_path`` and write the result back to ``target_path``, if the
        files have diverged according to :meth:`files_diverged`
        """
        source_file = File(source_path)
        target_file = File(target_path)

        if not await self.files_diverged(source_file, target_file):
            return TransformationResult(
                source=source_file, target=target_file, transformed=False
            )

        if self.config.cache:
            if await self._restore_from_cache(source_file, target_file):
                return TransformationResult(
                    source=source_file, target=target_file, transformed=True
                )

        await self.set_meta(source_file)

        if source_content := (await source_file.get_content()):
            transformed_content = await self.transform(source_file)
        else:
            transformed_content = source_content

        transformation_result = TransformationResult(
            source=source_file, target=target_file, transformed=False
        )

        if not self.config.check_only:
            if not (
                await target_file.exists()
                and await target_file.get_content() == transformed_content
            ):
                await target_file.parent.mkdir(parents=True, exist_ok=True)
                await target_file.write_text(transformed_content)
                transformation_result.transformed = True

            await self.set_meta(target_file)

        if self.config.cache:
            await self.cache.set(
                await self._cache_key_for(source_file), transformed_content
            )

        return transformation_result

    async def unasync_files(self) -> AsyncGenerator[TransformationResult, None]:
        """
        Transform all files specified in the configuration and return an async generator
        producing :class:`TransformationResult`.
        """
        if await self._config_changed():
            self.config.force_regen = True

        fs = []
        for source, target in self.config.files.items():
            fs.append(self.unasync_file(source, target))

        for c in asyncio.as_completed(fs):
            yield await c

        if self._meta:
            await self.cache.set("meta.json", msgspec.json.encode(self._meta).decode())

        await self.cache.set("__config_key__", self._config_key)


def unasync_files(config: Config) -> AsyncGenerator[TransformationResult, None]:
    """
    Transform all files specified in the configuration and return an async generator
    producing :class:`TransformationResult`.
    """
    env = Env(config=config)
    return env.unasync_files()



from __future__ import annotations

import dataclasses
import itertools
import re
import subprocess
import sys
from collections import defaultdict
from collections.abc import Iterable, Sequence
from typing import Any, Callable, TypeVar, Union, cast

import libcst as cst
import libcst.matchers as m
from libcst import MetadataWrapper
from libcst.metadata import QualifiedName, ScopeProvider
from libcst.native import parse_module

AnyImport = Union[cst.ImportFrom, cst.Import]
ScopedNodeT = TypeVar(
    "ScopedNodeT", bound=Union[cst.Module, cst.FunctionDef, cst.ClassDef]
)
AnyImportT = TypeVar("AnyImportT", bound=AnyImport)


NAME_REPLACEMENTS = {
    "__aenter__": "__enter__",
    "__aexit__": "__exit__",
    "__anext__": "__next__",
    "__aiter__": "__iter__",
    "athrow": "throw",
    "asend": "send",
    "anext": "next",
    "aiter": "iter",
    "StopAsyncIteration": "StopIteration",
    "contextlib.asynccontextmanager": "contextlib.contextmanager",
    "typing.AsyncIterable": "typing.Iterable",
    "collections.abc.AsyncIterable": "collections.abc.Iterable",
    "typing.AsyncIterator": "typing.Iterator",
    "collections.abc.AsyncIterator": "collections.abc.Iterator",
    "typing.AsyncGenerator": "typing.Generator",
    "collections.abc.AsyncGenerator": "collections.abc.Generator",
    "asyncio.sleep": "time.sleep",
    "anyio.sleep": "time.sleep",
    "anyio.Path": "pathlib.Path",
    "asyncio.Semaphore": "threading.Semaphore",
    "anyio.Semaphore": "threading.Semaphore",
    "asyncio.Lock": "threading.Lock",
    "anyio.Lock": "threading.Lock",
    "asyncio.Event": "threading.Event",
    "anyio.Event": "threading.Event",
    "asyncio.Barrier": "threading.Barrier",
}


UNASYNC_REPLACEMENTS = {
    "async def": "def",
    "async for": "for",
    "async with": "with",
    "await ": "",
    **NAME_REPLACEMENTS,
}


ASYNC_GEN_RE = re.compile(r"AsyncGenerator\[(.+,.+)]")


@dataclasses.dataclass
class TransformerMeta:
    needs_from_import: defaultdict[str, set[str]]
    needs_module_import: set[str]
    exclude: set[tuple[str, ...]]
    removed_names: defaultdict[str, set[str]]


@dataclasses.dataclass
class ImportMeta:
    module_imports: dict[str, cst.Import] = dataclasses.field(default_factory=dict)
    from_imports: dict[str, cst.ImportFrom] = dataclasses.field(default_factory=dict)


class StringTransformer:
    """A simple string based transformer"""

    def __init__(self, extra_name_replacements: dict[str, str] | None = None) -> None:
        self.replacements = {**UNASYNC_REPLACEMENTS, **(extra_name_replacements or {})}

    def __call__(self, source: str | bytes) -> str:
        if isinstance(source, bytes):
            source = source.decode()
        output = ASYNC_GEN_RE.sub(r"Generator[\1, None]", source)
        # Substitute 2 argument form of AsyncGenerator with the fully typed equivalent
        # 3 argument Generator form to pass mypy strict

        for string, replacement in self.replacements.items():
            output = output.replace(string, replacement)

        return output


def _create_name_or_attr(qualified_name: str) -> cst.Attribute | cst.Name:
    """Create a name or attribute from a fully qualified name"""
    if "." in qualified_name:
        attributes, qualified_name = qualified_name.rsplit(".", 1)
        # if attributes:
        return cst.Attribute(
            attr=cst.Name(value=qualified_name), value=_create_name_or_attr(attributes)
        )
    return cst.Name(value=qualified_name)


def _get_docstring_node(
    body: cst.BaseSuite | Sequence[cst.SimpleStatementLine | cst.BaseCompoundStatement],
) -> cst.SimpleString | cst.ConcatenatedString | None:
    """Get the ```SimpleString`` or ``ConcatenatedString`` node of ``body`` representing
    its docstring
    """
    expr: (
        cst.BaseSuite
        | cst.SimpleStatementLine
        | cst.BaseCompoundStatement
        | cst.BaseExpression
        | cst.BaseStatement
        | cst.BaseSmallStatement
    )
    if isinstance(body, Sequence):
        if body:
            expr = body[0]
        else:
            return None
    else:
        expr = body

    while isinstance(expr, (cst.BaseSuite, cst.SimpleStatementLine)):
        if len(expr.body) == 0:
            return None

        expr = expr.body[0]
    if not isinstance(expr, cst.Expr):
        return None

    val = expr.value
    if isinstance(val, (cst.SimpleString, cst.ConcatenatedString)):
        return val

    return None


def _get_full_name_for_import_from(node: cst.ImportFrom) -> str:
    return (
        "." * len(node.relative)
        + cst.helpers.get_full_name_for_node_or_raise(node.module)
        if node.module
        else ""
    )


def _create_import_from(module_name: str, names: Iterable[str]) -> cst.ImportFrom:
    relative: list[cst.Dot] = []
    if "." in module_name:
        orig_len = len(module_name)
        module_name = module_name.lstrip(".")
        relative = [cst.Dot() for _ in range(orig_len - len(module_name))]
    return cst.ImportFrom(
        module=_create_name_or_attr(module_name),
        names=[cst.ImportAlias(cst.Name(name)) for name in names],
        relative=relative,
    )


class TreeTransformer:
    """libcst based transformer"""

    def __init__(
        self,
        exclude: Iterable[str] | None = None,
        transform_docstrings: bool = True,
        extra_name_replacements: dict[str, str] | None = None,
        infer_type_checking_imports: bool = True,
        ruff_fix: bool = False,
        ruff_format: bool = False,
        file_name: str | None = None,
    ) -> None:
        self.exclude = exclude
        self.transform_docstrings = transform_docstrings
        self.name_replacements = {
            **NAME_REPLACEMENTS,
            **(extra_name_replacements or {}),
        }
        self.infer_type_checking_imports = infer_type_checking_imports
        self._post_transforms: list[Callable[[str, str, cst.Module], str]] = []
        self._file_name = file_name
        if ruff_fix:
            self._post_transforms.append(self._ruff_fix)
        if ruff_format:
            self._post_transforms.append(self._ruff_format)

    def _run_ruff(
        self, source: str, output: str, tree: cst.Module, cmd: list[str]
    ) -> str:
        args = [sys.executable, "-m", "ruff", *cmd, "--no-cache", "--quiet"]

        if self._file_name:
            args.append(f"--stdin-filename={self._file_name}")

        with subprocess.Popen(
            [*args, "-"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            encoding="utf-8",
        ) as process:
            stdout, stderr = process.communicate(input=output)
            if process.returncode != 0:
                raise ChildProcessError(f"Error calling ruff: {stderr}")
            return stdout

    def _ruff_fix(self, source: str, output: str, tree: cst.Module) -> str:
        return self._run_ruff(
            source, output, tree, cmd=["check", "--fix", "--fix-only"]
        )

    def _ruff_format(self, source: str, output: str, tree: cst.Module) -> str:
        return self._run_ruff(source, output, tree, cmd=["format"])

    def __call__(self, source: str) -> str:
        if not source:
            return ""

        wrapper = MetadataWrapper(parse_module(source), unsafe_skip_copy=True)
        result = wrapper.visit(
            _AsyncTransformer(
                exclude=self.exclude,
                name_replacements=self.name_replacements,
                transform_docstrings=self.transform_docstrings,
                infer_type_checking_imports=self.infer_type_checking_imports,
            )
        )
        output = result.code
        for post_transform in self._post_transforms:
            output = post_transform(source, output, result)

        return output


class _ReplaceAttributesMixin:
    """
    ``CSTTransformer`` mixin that replaces attributes by their fully qualified name
    """

    _replace_attributes: dict[str, dict[str, str]]

    def leave_Attribute(
        self, original_node: cst.Attribute, updated_node: cst.Attribute
    ) -> cst.BaseExpression:
        maybe_name = updated_node.value
        if not isinstance(maybe_name, cst.Name):
            return updated_node
        name = maybe_name.value

        if not (attribute_replacements := self._replace_attributes.get(name)):
            return updated_node

        attr_name = updated_node.attr.value

        if attr_replacement := attribute_replacements.get(attr_name):
            updated_node = updated_node.with_changes(
                attr=_create_name_or_attr(attr_replacement)
            )

        return updated_node


class _ReplaceNamesMixin:
    """``CSTTransformer`` mixin that replaces names by their fully qualified name"""

    _in_import: bool = False
    _name_replacements: dict[str, str]
    meta: TransformerMeta

    def _create_replacement_name_or_attribute_for(
        self, node: cst.Name | cst.Attribute, replace_with: str
    ) -> cst.Name | cst.Attribute:
        """Create a replacement name or attribute for the given note"""
        if isinstance(node, cst.Name):
            if "." in replace_with:
                module_name, attr = replace_with.rsplit(".", 1)
                self.meta.needs_from_import[module_name].add(attr)
                return cst.Name(attr)
            return cst.Name(replace_with)

        self.meta.needs_module_import.add(replace_with.rsplit(".", 1)[0])
        return _create_name_or_attr(replace_with)

    def leave_Name(
        self, original_node: cst.Name, updated_node: cst.Name
    ) -> cst.BaseExpression:
        if self._in_import:
            return updated_node

        old_name: str = updated_node.value
        new_name: str | None = None

        try:
            scope = self.get_metadata(  # type: ignore[attr-defined]
                ScopeProvider,
                original_node,
            )
            qualified_names: set[QualifiedName] = scope.get_qualified_names_for(
                original_node
            )

            for _name in qualified_names:
                new_name = self._name_replacements.get(_name.name)
                if new_name:
                    old_name = _name.name
                    break
            if new_name and len(qualified_names) > 1:
                raise RuntimeError("Could not determine fully qualified name")

        except KeyError:
            pass

        if not new_name:
            new_name = self._name_replacements.get(old_name)

        if new_name:
            if (
                "." in old_name
            ):  # if not a fully qualified name we don't care about imports
                module_name, attr = old_name.rsplit(".", 1)
                self.meta.removed_names[module_name].add(attr)

            return self._create_replacement_name_or_attribute_for(
                updated_node, new_name
            )

        return updated_node


class ContextManagerTransformer(cst.CSTTransformer):
    def __init__(self, *, asname: str, meta: TransformerMeta) -> None:
        super().__init__()
        self.name = asname
        self.meta = meta


class _AsyncExitStackBodyTransformer(
    _ReplaceAttributesMixin, ContextManagerTransformer
):
    """
    ``CSTTransformer`` transforming a ``contextlib.AsyncExitStack into a
    ``contextlib.ExitStack``
    """

    def __init__(self, *, asname: str, meta: TransformerMeta) -> None:
        super().__init__(asname=asname, meta=meta)
        self._replace_attributes = {
            self.name: {
                "enter_async_context": "enter_context",
                "push_async_exit": "push",
                "push_async_callback": "callback",
                "aclose": "close",
            }
        }


class _AnyioTaskGroupBodyTransformer(
    _ReplaceAttributesMixin, _ReplaceNamesMixin, ContextManagerTransformer
):
    """
    ``CSTTransformer`` transforming ``anyio.create_task_group()`` into usages of
    ``concurrent.futures.ThreadPoolExecutor``
    """

    def __init__(self, *, asname: str, meta: TransformerMeta) -> None:
        self._name_replacements = {}
        if asname in {"task_group", "tg"}:
            self._name_replacements["task_group"] = "executor"
            asname = "executor"
        self._replace_attributes = {asname: {"start_soon": "submit", "start": "submit"}}

        super().__init__(asname=asname, meta=meta)


class _TaskGroupBodyTransformer(
    _ReplaceAttributesMixin, _ReplaceNamesMixin, ContextManagerTransformer
):
    """
    ``CSTTransformer`` transforming ``asyncio.TaskGroup()`` into usages of
    ``concurrent.futures.ThreadPoolExecutor``
    """

    def __init__(self, *, asname: str, meta: TransformerMeta) -> None:
        super().__init__(asname=asname, meta=meta)
        self.call_matcher = m.Call(
            func=m.Attribute(
                value=m.Name(value=self.name), attr=m.Name(value="create_task")
            )
        )

        self._name_replacements = {}

        if self.name in {"task_group", "tg"}:
            self._name_replacements["task_group"] = "executor"
            self.name = "executor"

        self._replace_attributes = {self.name: {"create_task": "submit"}}

    def leave_Call(
        self, original_node: cst.Call, updated_node: cst.Call
    ) -> cst.BaseExpression:
        if not m.matches(original_node, self.call_matcher):
            return updated_node

        first_arg = updated_node.args[0]
        if not m.matches(first_arg, m.Arg(value=m.Call(m.DoNotCare()))):
            return updated_node

        call = cst.ensure_type(first_arg.value, cst.Call)
        new_arg = first_arg.with_changes(value=call.func)
        return updated_node.with_changes(args=[new_arg, *call.args])


class _AsyncTransformer(_ReplaceNamesMixin, cst.CSTTransformer):
    """Main entry point for unasyncd"""

    METADATA_DEPENDENCIES = (ScopeProvider,)

    def __init__(
        self,
        *,
        exclude: Iterable[str] | None = None,
        name_replacements: dict[str, str],
        transform_docstrings: bool,
        infer_type_checking_imports: bool = True,
    ):
        """

        :param exclude: Iterable of fully qualified names of classes, methods and
            functions to exclude
        :param name_replacements: Mapping of fully qualified names to their replacements
        :param transform_docstrings: If ``True``, transform docstrings
        """
        super().__init__()
        self._should_transform_docstrings = transform_docstrings
        self._should_infer_type_checking_imports = infer_type_checking_imports
        self._name_replacements = name_replacements
        self._attribute_replacements = {
            name: replacement
            for name, replacement in name_replacements.items()
            if "." in name
        }

        self._scoped_node_imports: defaultdict[cst.CSTNode, ImportMeta] = defaultdict(
            lambda: ImportMeta(module_imports={}, from_imports={})
        )

        self._expressions_to_remove: set[cst.Expr] = set()
        self._string_transformer = StringTransformer(self._name_replacements)

        self._current_scope_name: tuple[str, ...] = tuple()
        self._scope_nodes: list[cst.CSTNode] = []

        self._if_type_checking_node: cst.If | None = None
        self._if_type_checking_imports: set[AnyImport] = set()
        self._in_if_type_checking_block: bool = False
        self._in_import = False

        self.meta = TransformerMeta(
            needs_from_import=defaultdict(set),
            needs_module_import=set(),
            removed_names=defaultdict(set),
            exclude={tuple(p.split(".")) for p in exclude or []},
        )

    def visit_If(self, node: cst.If) -> bool | None:
        if (
            isinstance(node.test, cst.Name)
            and self.get_qualified_name(node.test) == "typing.TYPE_CHECKING"
        ):
            self._if_type_checking_node = node
            self._in_if_type_checking_block = True

        return None

    def leave_If(self, original_node: cst.If, updated_node: cst.If) -> Any:
        if original_node is self._if_type_checking_node:
            self._in_if_type_checking_block = False

            # replace with the updated node, so we can check for identity later on when
            # performing transformations
            self._if_type_checking_node = updated_node

        return updated_node

    @property
    def current_scope_node(self) -> cst.CSTNode | None:
        """The parent node of the current scope (excludes comprehension scope)"""
        if self._scope_nodes:
            return self._scope_nodes[-1]
        return None

    @property
    def _should_transform_current_node(self) -> bool:
        return self._current_scope_name not in self.meta.exclude

    def get_qualified_name(self, node: cst.CSTNode) -> str | None:
        """Get the fully qualified name of ``node``, relative to the current module."""
        try:
            if not (scope := self.get_metadata(ScopeProvider, node)):
                return None

            qualified_names: set[QualifiedName] = set(
                scope.get_qualified_names_for(node)
            )

            if len(qualified_names) == 1:
                return qualified_names.pop().name
            elif qualified_names:  # multiple names found
                raise ValueError(
                    f"Could not determine fully qualified name for {node!r}"
                )

        except KeyError:
            pass

        return None

    def visit_Module(self, node: cst.Module) -> bool:
        self._scope_nodes.append(node)
        return True

    def visit_ClassDef(self, node: cst.ClassDef) -> bool | None:
        self._current_scope_name = (*self._current_scope_name, node.name.value)
        self._scope_nodes.append(node)
        return self._should_transform_current_node

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> (
        cst.BaseStatement | cst.FlattenSentinel[cst.BaseStatement] | cst.RemovalSentinel
    ):
        self._current_scope_name = self._current_scope_name[:-1]
        self._scope_nodes = self._scope_nodes[:-1]
        if self._should_transform_docstrings and self._should_transform_current_node:
            updated_node = self._transform_docstring(updated_node)
        return updated_node

    def visit_FunctionDef(self, node: cst.FunctionDef) -> bool | None:
        self._current_scope_name = (*self._current_scope_name, node.name.value)
        self._scope_nodes.append(node)
        return self._should_transform_current_node

    def leave_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> (
        cst.BaseStatement | cst.FlattenSentinel[cst.BaseStatement] | cst.RemovalSentinel
    ):
        if self._should_transform_current_node:
            if updated_node.asynchronous:
                updated_node = updated_node.with_changes(asynchronous=None)

            if self._should_transform_docstrings:
                updated_node = self._transform_docstring(updated_node)

        self._current_scope_name = self._current_scope_name[:-1]
        self._scope_nodes = self._scope_nodes[:-1]
        return updated_node

    def _transform_docstring(self, updated_node: ScopedNodeT) -> ScopedNodeT:
        if not (docstring_node := _get_docstring_node(updated_node.body)):
            return updated_node

        if not (original_docstring := docstring_node.evaluated_value):
            return updated_node

        updated_docstring = self._string_transformer(original_docstring)
        quote = (
            docstring_node.quote
            if isinstance(docstring_node, cst.SimpleString)
            else docstring_node.left.quote
        )

        if updated_docstring != original_docstring:
            updated_node = updated_node.with_deep_changes(  # type: ignore[assignment]
                docstring_node,
                value=quote + updated_docstring + quote,
            )

        return updated_node

    def _register_import(self, node: cst.Import | cst.ImportFrom) -> None:
        if not self.current_scope_node:
            return

        node_imports = self._scoped_node_imports[self.current_scope_node]
        if isinstance(node, cst.Import):
            for name in node.names:
                node_imports.module_imports[name.evaluated_name] = node
            return None

        module_name = _get_full_name_for_import_from(node)
        node_imports.from_imports[module_name] = node

    def _leave_any_import(
        self, original_node: AnyImport, updated_node: AnyImport
    ) -> (
        cst.BaseSmallStatement
        | cst.FlattenSentinel[cst.BaseSmallStatement]
        | cst.RemovalSentinel
    ):
        self._in_import = False

        self._register_import(updated_node)
        if self._in_if_type_checking_block:
            self._if_type_checking_imports.add(updated_node)

        return updated_node

    def visit_ImportFrom(self, node: cst.ImportFrom) -> bool:
        self._in_import = True
        return True

    def visit_Import(self, node: cst.Import) -> bool | None:
        self._in_import = True
        return None

    def leave_ImportFrom(
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> (
        cst.BaseSmallStatement
        | cst.FlattenSentinel[cst.BaseSmallStatement]
        | cst.RemovalSentinel
    ):
        return self._leave_any_import(original_node, updated_node)

    def leave_Import(
        self, original_node: cst.Import, updated_node: cst.Import
    ) -> (
        cst.BaseSmallStatement
        | cst.FlattenSentinel[cst.BaseSmallStatement]
        | cst.RemovalSentinel
    ):
        return self._leave_any_import(original_node, updated_node)

    def leave_Attribute(
        self, original_node: cst.Attribute, updated_node: cst.Attribute
    ) -> cst.BaseExpression:
        if not (qualified_name := self.get_qualified_name(original_node)):
            return updated_node

        if not (replacement := self._attribute_replacements.get(qualified_name)):
            return updated_node

        module_name, attr = qualified_name.rsplit(".", 1)
        self.meta.removed_names[module_name].add(attr)
        self.meta.needs_module_import.add(replacement.rsplit(".", 1)[0])

        return _create_name_or_attr(replacement)

    def _transform_context_manager(
        self,
        *,
        updated_node: cst.With,
        with_item: cst.WithItem,
        item_call_func: cst.Name | cst.Attribute,
        visitor_class: type[ContextManagerTransformer],
        replace_call: str,
    ) -> cst.With:
        """Transform a context manager's with-item"""
        if not with_item.asname:
            raise ValueError(f"Missing as-name for context manager item {with_item}")

        new_call_func = self._create_replacement_name_or_attribute_for(
            item_call_func, replace_call
        )

        updated_node = updated_node.with_deep_changes(
            with_item, item=cst.Call(func=new_call_func)
        )
        asname = cst.ensure_type(with_item.asname.name, cst.Name)  # can be a tuple,
        # but in none of the supported cases, so we ensure its type here
        visitor = visitor_class(asname=asname.value, meta=self.meta)
        visitor.get_metadata = self.get_metadata  # type: ignore[method-assign]
        return cst.ensure_type(updated_node.visit(visitor), cst.With)

    def _transform_context_managers(
        self,
        original_node: cst.With,
        updated_node: cst.With,
        with_item: cst.WithItem,
    ) -> cst.With:
        """Transform asynchronous context managers that need special handling, such as
        ``asyncio.TaskGroup``
        """
        if not m.matches(
            with_item, m.WithItem(m.Call(func=m.OneOf(m.Name(), m.Attribute())))
        ):
            return updated_node

        item_call_func: cst.Attribute | cst.Name = with_item.item.func  # type: ignore[attr-defined]

        if not (scope := self.get_metadata(ScopeProvider, original_node)):
            raise RuntimeError("Scope not found")

        qualified_names = set(scope.get_qualified_names_for(item_call_func))

        for name in qualified_names:
            if name.name == "contextlib.AsyncExitStack":
                self.meta.removed_names["contextlib"].add("AsyncExitStack")
                return self._transform_context_manager(
                    updated_node=updated_node,
                    with_item=with_item,
                    replace_call="contextlib.ExitStack",
                    visitor_class=_AsyncExitStackBodyTransformer,
                    item_call_func=item_call_func,
                )

            if name.name == "asyncio.TaskGroup":
                self.meta.removed_names["asyncio"].add("TaskGroup")
                return self._transform_context_manager(
                    updated_node=updated_node,
                    with_item=with_item,
                    replace_call="concurrent.futures.ThreadPoolExecutor",
                    visitor_class=_TaskGroupBodyTransformer,
                    item_call_func=item_call_func,
                )

            if name.name == "anyio.create_task_group":
                self.meta.removed_names["anyio"].add("create_task_group")
                return self._transform_context_manager(
                    updated_node=updated_node,
                    with_item=with_item,
                    replace_call="concurrent.futures.ThreadPoolExecutor",
                    visitor_class=_AnyioTaskGroupBodyTransformer,
                    item_call_func=item_call_func,
                )

        return updated_node

    def leave_With(
        self, original_node: cst.With, updated_node: cst.With
    ) -> (
        cst.BaseStatement | cst.FlattenSentinel[cst.BaseStatement] | cst.RemovalSentinel
    ):
        if not updated_node.asynchronous:
            return updated_node

        for with_item in updated_node.items:
            updated_node = self._transform_context_managers(
                original_node, updated_node, with_item
            )
        return updated_node.with_changes(asynchronous=None)

    def visit_Expr(self, node: cst.Expr) -> bool | None:
        """Register expressions to be removed entirely. Currently removes
        ``await asyncio.sleep(0)`` and ``await anyio.sleep(0)``
        """
        if m.matches(node, m.Expr(m.Await(expression=m.Call(m.DoNotCare())))):
            call = cst.ensure_type(node.value, cst.Await).expression
            qualified_name = self.get_qualified_name(call)
            if qualified_name not in {"asyncio.sleep", "anyio.sleep"}:
                return True

            if m.matches(
                call, m.Call(args=[m.AtLeastN(m.Arg(m.Integer(value="0")), n=1)])
            ):
                self._expressions_to_remove.add(node)
                removed_mod, removed_name = qualified_name.split(".")
                self.meta.removed_names[removed_mod].add(removed_name)
                return False
        return True

    def leave_Expr(
        self, original_node: cst.Expr, updated_node: cst.Expr
    ) -> cst.BaseSmallStatement | cst.RemovalSentinel:
        """Remove expressions registered to be removed"""
        if original_node in self._expressions_to_remove:
            return cst.RemoveFromParent()
        return updated_node

    def leave_Await(
        self, original_node: cst.Await, updated_node: cst.Await
    ) -> cst.BaseExpression | Any:
        """Unwrap await: ``await foo`` > ``foo``"""
        return updated_node.expression.with_changes(
            lpar=updated_node.lpar,
            rpar=updated_node.rpar,
        )

    def leave_For(
        self, original_node: cst.For, updated_node: cst.For
    ) -> (
        cst.BaseStatement | cst.FlattenSentinel[cst.BaseStatement] | cst.RemovalSentinel
    ):
        """Replace ``async for i in thing`` > ``for i in thing"""
        if updated_node.asynchronous:
            updated_node = updated_node.with_changes(asynchronous=None)
        return updated_node

    def leave_CompFor(
        self, original_node: cst.CompFor, updated_node: cst.CompFor
    ) -> cst.CompFor:
        if updated_node.asynchronous:
            updated_node = updated_node.with_changes(asynchronous=None)
        return updated_node

    def leave_Subscript(
        self, original_node: cst.Subscript, updated_node: cst.Subscript
    ) -> cst.BaseExpression:
        """Fix ``typing.Generator`` generic form by adding a third ``None`` as a last
        subscript element: ``typing.AsyncGenerator[None, None]`` >
        ``typing.Generator[None, None, None]``.
        """
        qualified_name = self.get_qualified_name(original_node)
        if qualified_name in (
            "typing.AsyncGenerator",
            "collections.abc.AsyncGenerator",
        ):
            return self._fix_async_generator(updated_node)

        if qualified_name in ("typing.Awaitable", "collections.abc.Awaitable"):
            return self._transform_awaitable(updated_node)

        return updated_node

    def _transform_awaitable(self, updated_node: cst.Subscript) -> cst.BaseExpression:
        if len(updated_node.slice) > 1:
            raise ValueError("Too many elements in typing.Awaitable")
        index = updated_node.slice[0].slice
        if not isinstance(index, cst.Index):
            raise ValueError(f"Expected index element. Got {type(index)}")
        return index.value

    def _fix_async_generator(self, updated_node: cst.Subscript) -> cst.Subscript:
        none_element = cst.SubscriptElement(slice=cst.Index(cst.Name("None")))
        updated_node = updated_node.with_changes(
            slice=[*updated_node.slice, none_element]
        )
        return updated_node

    def leave_Assign(
        self, original_node: cst.Assign, updated_node: cst.Assign
    ) -> (
        cst.BaseSmallStatement
        | cst.FlattenSentinel[cst.BaseSmallStatement]
        | cst.RemovalSentinel
    ):
        """Handle ``__all__`` and its imports"""

        if not isinstance(self.current_scope_node, cst.Module):
            return updated_node

        if not m.matches(
            updated_node, m.Assign(targets=[m.AssignTarget(target=m.Name("__all__"))])
        ):
            return updated_node

        imported_names: dict[str, str] = {}
        for module_name, import_ in self._scoped_node_imports[
            self.current_scope_node
        ].from_imports.items():
            if isinstance(import_.names, cst.ImportStar):
                continue

            for name in import_.names:
                full_name = name.evaluated_alias or name.evaluated_name
                imported_names[full_name] = (
                    f"{_get_full_name_for_import_from(import_)}.{full_name}"
                )

        elements: list[cst.Element] = updated_node.value.elements  # type: ignore[attr-defined]
        updated_elements: list[cst.Element] = []
        for element in elements:
            if not isinstance(element.value, cst.SimpleString):
                continue

            value = cast(str, element.value.evaluated_value)
            if not (qualified_name := imported_names.get(value)):
                continue

            if not (replacement := self._name_replacements.get(qualified_name)):
                continue

            if "." in replacement:
                module_name, attr = replacement.rsplit(".", 1)
                self.meta.needs_from_import[module_name].add(attr)
                replacement = attr
            else:
                self.meta.needs_module_import.add(replacement.rsplit(".", 1)[0])

            quote = element.value.quote

            updated_elements.append(
                element.with_deep_changes(
                    element.value, value=f"{quote}{replacement}{quote}"
                )
            )

        return updated_node.with_deep_changes(
            updated_node.value, elements=updated_elements
        )

    def _find_first_non_import_line(self, updated_node: cst.Module) -> int:
        """Get the index of the first line of the module that is not an import,
        skipping module level docstrings and comments.
        """
        import_matcher = m.SimpleStatementLine(
            body=[m.AtLeastN(m.OneOf(m.Import(), m.ImportFrom()), n=1)]
        )
        string_matcher = m.SimpleStatementLine([m.Expr(m.SimpleString())])
        for i, node in enumerate(updated_node.body):
            if m.matches(node, string_matcher):
                continue
            if not m.matches(node, import_matcher):
                return i
        return 0

    def _extract_if_type_checking_imports(
        self, existing_imports: dict[str, AnyImportT]
    ) -> dict[str, set[str]]:
        type_checking_imports: dict[str, set[str]] = defaultdict(set)
        for module_name, import_ in existing_imports.items():
            if import_ not in self._if_type_checking_imports:
                continue
            if not isinstance(import_.names, cst.ImportStar):
                type_checking_imports[module_name].update(
                    [n.evaluated_alias or n.evaluated_name for n in import_.names]
                )
        return type_checking_imports

    def _create_from_imports(
        self,
        existing_from_imports: dict[str, cst.ImportFrom],
        replaced_names: dict[str, dict[str, str]],
        new_type_checking_imports: set[AnyImport],
    ) -> tuple[dict[cst.ImportFrom, set[str]], list[cst.ImportFrom]]:
        """Return a tuple of ``FromImport``s to update and ``FromImport``s to add"""
        from_imports_to_update: defaultdict[cst.ImportFrom, set[str]] = defaultdict(set)
        from_imports_to_add: defaultdict[str, set[str]] = defaultdict(set)
        existing_if_type_checking_imports = self._extract_if_type_checking_imports(
            existing_from_imports
        )

        for module_name, names in self.meta.needs_from_import.items():
            if _import := existing_from_imports.get(module_name):
                if isinstance(_import.names, cst.ImportStar):
                    continue
                existing_names = {
                    alias.evaluated_alias or alias.evaluated_name
                    for alias in _import.names
                }
                from_imports_to_update[_import].update(names - existing_names)
            else:
                from_imports_to_add[module_name].update(names)

        new_imports = []
        for module_name, names in from_imports_to_add.items():
            new_import_names = set()
            new_type_checking_import_names = set()

            if replaced_imports := replaced_names.get(module_name):
                # we have replaced names, so we check for the ones that exist
                # exclusively within an if TYPE_CHECKING block to ensure we only add
                # those there
                for new_name in names:
                    original_fq_name = replaced_imports[new_name]
                    orig_module_name, orig_name = original_fq_name.rsplit(".", 1)
                    if orig_name in existing_if_type_checking_imports.get(
                        orig_module_name, []
                    ):
                        new_type_checking_import_names.add(new_name)
                    else:
                        new_import_names.add(new_name)
            else:
                new_import_names.update(names)

            if new_import_names:
                new_imports.append(_create_import_from(module_name, new_import_names))

            if new_type_checking_import_names:
                new_type_checking_imports.add(
                    _create_import_from(module_name, new_type_checking_import_names)
                )

        return from_imports_to_update, new_imports

    def _create_module_imports(
        self,
        existing_module_imports: dict[str, cst.Import],
        replaced_names: dict[str, dict[str, str]],
        type_checking_imports: set[AnyImport],
    ) -> set[cst.Import]:
        """Return a list of ``Import``s to add"""
        new_imports: set[cst.Import] = set()
        existing_tc_import = self._extract_if_type_checking_imports(
            existing_module_imports
        )
        for module_name in self.meta.needs_module_import:
            if _import := existing_module_imports.get(module_name):
                continue

            if isinstance(_import, cst.ImportFrom):
                continue

            new_import = cst.Import(
                names=[cst.ImportAlias(name=_create_name_or_attr(module_name))]
            )

            maybe_replaced_imports = replaced_names.get(module_name)
            if not maybe_replaced_imports:
                new_imports.add(new_import)
                continue

            for new_name, original_fq_name in maybe_replaced_imports.items():
                orig_module_name, _ = original_fq_name.rsplit(".", 1)
                if orig_module_name in existing_tc_import:
                    type_checking_imports.add(new_import)
                else:
                    new_imports.add(new_import)

        return new_imports

    def _fix_module_imports(
        self,
        *,
        node: cst.Module | MetadataWrapper,
        from_imports: dict[str, cst.ImportFrom],
        module_imports: dict[str, cst.Import],
    ) -> cst.Module:
        """Modify a module's imports.

        - Update existing imports with names that have been requested by a
          transformation
        - Add new imports that have not been covered in the previous step
        - Remove names from imports specified in ``imports_to_remove``. If an import has
          no names left, it is removed entirely
        """
        maybe_replaced_names: dict[str, dict[str, str]] = defaultdict(dict)
        add_to_typechecking_import: set[AnyImport] = set()

        for name, replacement in self._name_replacements.items():
            if "." not in name:
                continue
            module_name, replacement = replacement.rsplit(".", 1)
            maybe_replaced_names[module_name][replacement] = name

        from_imports_to_update, new_from_imports = self._create_from_imports(
            from_imports,
            replaced_names=maybe_replaced_names,
            new_type_checking_imports=add_to_typechecking_import,
        )
        new_module_imports = self._create_module_imports(
            module_imports,
            replaced_names=maybe_replaced_names,
            type_checking_imports=add_to_typechecking_import,
        )

        imports_to_add: Iterable[AnyImport] = itertools.chain(
            new_from_imports, new_module_imports
        )

        if not self._should_infer_type_checking_imports:
            imports_to_add = itertools.chain(imports_to_add, add_to_typechecking_import)
            add_to_typechecking_import = set()

        updated_node = node.visit(
            _ImportTransformer(
                imports_to_update=from_imports_to_update,
                if_type_checking_node=self._if_type_checking_node,
                type_checking_imports_to_add=add_to_typechecking_import,
            )
        )

        import_offset = self._find_first_non_import_line(updated_node)

        new_module_body = [
            *updated_node.body[:import_offset],
            *[cst.SimpleStatementLine([new_import]) for new_import in imports_to_add],
            *updated_node.body[import_offset:],
        ]

        updated_node = updated_node.with_changes(body=new_module_body)

        return updated_node

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        import_meta = self._scoped_node_imports[original_node]
        updated_node = self._fix_module_imports(
            node=updated_node,
            module_imports=import_meta.module_imports,
            from_imports=import_meta.from_imports,
        )

        if self._should_transform_docstrings and self._should_transform_current_node:
            updated_node = self._transform_docstring(updated_node)

        return updated_node


class _ImportTransformer(cst.CSTTransformer):
    def __init__(
        self,
        *,
        imports_to_update: dict[cst.ImportFrom, set[str]],
        if_type_checking_node: cst.If | None,
        type_checking_imports_to_add: Iterable[AnyImport],
    ) -> None:
        super().__init__()
        self.imports_to_update = imports_to_update
        self.if_type_checking_node = if_type_checking_node
        self.type_checking_imports_to_add = type_checking_imports_to_add

    def leave_If(
        self, original_node: cst.If, updated_node: cst.If
    ) -> (
        cst.BaseStatement | cst.FlattenSentinel[cst.BaseStatement] | cst.RemovalSentinel
    ):
        if not self.if_type_checking_node:
            return updated_node

        if original_node is not self.if_type_checking_node:
            return updated_node

        return updated_node.with_deep_changes(
            updated_node.body,
            body=[
                *updated_node.body.body,
                *[
                    cst.SimpleStatementLine([import_])
                    for import_ in self.type_checking_imports_to_add
                ],
            ],
        )

    def leave_ImportFrom(
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> cst.ImportFrom | cst.RemovalSentinel:
        aliases_to_add = self.imports_to_update.get(original_node)
        if aliases_to_add:
            new_aliases = [cst.ImportAlias(cst.Name(name)) for name in aliases_to_add]

            current_names = (
                updated_node.names
                if not isinstance(updated_node.names, cst.ImportStar)
                else [updated_node.names]  # type: ignore[list-item]
            )

            updated_node = updated_node.with_changes(
                names=[*current_names, *new_aliases]
            )

        return updated_node

    # since we're only interested in module level imports, we don't need to visit
    # class and function definitions

    def visit_ClassDef(self, node: cst.ClassDef) -> bool:
        return False

    def visit_FunctionDef(self, node: cst.FunctionDef) -> bool:
        return False



from __future__ import annotations

import hashlib
import itertools
from pathlib import Path
from typing import Any

import msgspec

from ._version import VERSION


class Config(msgspec.Struct):
    add_editors_note: bool
    transform_docstrings: bool
    cache: bool
    extra_replacements: dict[str, dict[str, str]]
    exclude: dict[str, list[str]]
    files: dict[str, str]
    force_regen: bool
    check_only: bool
    infer_type_checking_imports: bool
    ruff_fix: bool
    ruff_format: bool

    def key(self) -> str:
        return hashlib.sha1(msgspec.json.encode(self) + VERSION.encode()).hexdigest()


def _collect_paths(file_names: dict[str, str]) -> dict[str, str]:
    files = {}
    for source_name, target_name in file_names.items():
        source_path = Path(source_name)
        target_path = Path(target_name)

        if not source_path.exists():
            raise FileNotFoundError(str(source_path))

        if source_path.is_dir():
            for file in source_path.rglob("*.py"):
                files[str(file)] = str(target_path / (file.relative_to(source_path)))
        else:
            files[str(source_path)] = str(target_path)

    return files


def load_config(path: Path | None, **defaults: Any) -> Config:
    raw_config = (
        msgspec.toml.decode(path.read_bytes()).get("tool", {}).get("unasyncd", {})
        if path
        else {}
    )

    raw_config.update({k: v for k, v in defaults.items() if v is not None})

    exclude = raw_config.get("exclude", [])
    per_file_exclude = raw_config.get("per_file_exclude", {})

    extra_replacements = raw_config.get("add_replacements", {})
    per_file_extra_replacements = raw_config.get("per_file_add_replacements", {})

    files = _collect_paths(raw_config.get("files", {}))

    return Config(
        files=files,
        extra_replacements={
            file: {**extra_replacements, **per_file_extra_replacements.get(file, {})}
            for file in itertools.chain(files, per_file_extra_replacements)
        },
        exclude={
            file: [*exclude, *per_file_exclude.get(file, [])]
            for file in itertools.chain(files, per_file_exclude)
        },
        add_editors_note=raw_config.get("add_editors_note", False),
        transform_docstrings=raw_config.get("transform_docstrings", False),
        cache=raw_config.get("cache", True),
        check_only=raw_config.get("check_only", False),
        force_regen=raw_config.get("force_regen", False),
        infer_type_checking_imports=raw_config.get("infer_type_checking_imports", True),
        ruff_fix=raw_config.get("ruff_fix", False),
        ruff_format=raw_config.get("ruff_format", False),
    )



from textwrap import dedent

import libcst as cst
import pytest
import tomli_w

from unasyncd.transformers import TreeTransformer


@pytest.fixture
def transformer() -> TreeTransformer:
    return TreeTransformer()


def test_unwrap_name_or_attribute() -> None:
    node = cst.Attribute(
        value=cst.Attribute(value=cst.Name("foo"), attr=cst.Name("bar")),
        attr=cst.Name("baz"),
    )
    assert cst.helpers.get_full_name_for_node_or_raise(node) == "foo.bar.baz"


def test_async_def(transformer: TreeTransformer) -> None:
    source = """
    async def foo():
        return None
    """

    expected = """
    def foo():
        return None
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_await(transformer: TreeTransformer) -> None:
    assert transformer("await foo") == "foo"
    assert transformer("await foo()") == "foo()"


def test_await_in_function(transformer: TreeTransformer) -> None:
    source = """
    async def foo():
        return await foo()
    """

    expected = """
    def foo():
        return foo()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_for(transformer: TreeTransformer) -> None:
    source = """
    async for i in thing():
      pass
    """

    expected = """
    for i in thing():
      pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_for_with_await(transformer: TreeTransformer) -> None:
    source = """
    async for i in await thing():
      pass
    """

    expected = """
    for i in thing():
      pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_with(transformer: TreeTransformer) -> None:
    source = """
    async with foo() as bar:
      pass
    """

    expected = """
    with foo() as bar:
      pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_method(transformer: TreeTransformer) -> None:
    source = """
    class Foo:
        async def bar(self):
            pass
    """

    expected = """
    class Foo:
        def bar(self):
            pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_magic_method_replacement(transformer: TreeTransformer) -> None:
    source = """
    class Foo:
        async def __aenter__():
            pass

        async def __aexit__():
            pass

        async def __anext__():
            pass

        async def __aiter__():
            pass

        async def athrow():
            pass

        async def asend():
            pass
    """

    expected = """
    class Foo:
        def __enter__():
            pass

        def __exit__():
            pass

        def __next__():
            pass

        def __iter__():
            pass

        def throw():
            pass

        def send():
            pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_athrow(transformer: TreeTransformer) -> None:
    source = "await x.athrow()"
    expected = "x.throw()"

    assert transformer(source) == expected


def test_asend(transformer: TreeTransformer) -> None:
    source = "await x.asend()"
    expected = "x.send()"

    assert transformer(source) == expected


def test_anext(transformer: TreeTransformer) -> None:
    source = "anext(foo)"
    expected = "next(foo)"

    assert transformer(source) == expected


def test_aiter(transformer: TreeTransformer) -> None:
    source = "aiter(foo)"
    expected = "iter(foo)"

    assert transformer(source) == expected


def test_stop_async_iteration(transformer: TreeTransformer) -> None:
    source = """
    async def foo():
        yield
        raise StopAsyncIteration
    """

    expected = """
    def foo():
        yield
        raise StopIteration
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asynccontextmanager(transformer: TreeTransformer) -> None:
    source = """
    from contextlib import asynccontextmanager

    @asynccontextmanager
    async def foo():
        yield
    """

    expected = """
    from contextlib import asynccontextmanager, contextmanager

    @contextmanager
    def foo():
        yield
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asynccontextmanager_alias_import(transformer: TreeTransformer) -> None:
    source = """
    from contextlib import asynccontextmanager as something_else

    @something_else
    async def foo():
        yield
    """

    expected = """
    from contextlib import asynccontextmanager as something_else, contextmanager

    @contextmanager
    def foo():
        yield
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asynccontextmanager_not_from_contextlib_ignore(
    transformer: TreeTransformer,
) -> None:
    source = """
    async def asynccontextmanager():
        pass
    """

    expected = """
    def asynccontextmanager():
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asynccontextmanager_module_import(transformer: TreeTransformer) -> None:
    source = """
    import contextlib

    @contextlib.asynccontextmanager
    async def foo():
        yield
    """

    expected = """
    import contextlib

    @contextlib.contextmanager
    def foo():
        yield
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_generator(transformer: TreeTransformer) -> None:
    source = """
    from typing import AsyncGenerator
    x: AsyncGenerator
    """

    expected = """
    from typing import AsyncGenerator, Generator
    x: Generator
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_generator_collections_abc(transformer: TreeTransformer) -> None:
    source = """
    from collections.abc import AsyncGenerator
    x: AsyncGenerator
    """

    expected = """
    from collections.abc import AsyncGenerator, Generator
    x: Generator
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_generator_annotation(transformer: TreeTransformer) -> None:
    source = """
    from typing import AsyncGenerator
    x: AsyncGenerator[str, int]
    """

    expected = """
    from typing import AsyncGenerator, Generator
    x: Generator[str, int, None]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_generator_annotation_collections_abc(
    transformer: TreeTransformer,
) -> None:
    source = """
    from collections.abc import AsyncGenerator
    x: AsyncGenerator[str, int]
    """

    expected = """
    from collections.abc import AsyncGenerator, Generator
    x: Generator[str, int, None]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_generator_annotation_typing_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import typing
    x: typing.AsyncGenerator[str, int]
    """

    expected = """
    import typing
    x: typing.Generator[str, int, None]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_generator_class(transformer: TreeTransformer) -> None:
    source = """
    from typing import AsyncGenerator

    class Foo(AsyncGenerator[str, int]):
        pass
    """

    expected = """
    from typing import AsyncGenerator, Generator

    class Foo(Generator[str, int, None]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_generator_class_collections_abc(transformer: TreeTransformer) -> None:
    source = """
    from collections.abc import AsyncGenerator

    class Foo(AsyncGenerator[str, int]):
        pass
    """

    expected = """
    from collections.abc import AsyncGenerator, Generator

    class Foo(Generator[str, int, None]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_generator_class_typing_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import typing

    class Foo(typing.AsyncGenerator[str, int]):
        pass
    """

    expected = """
    import typing

    class Foo(typing.Generator[str, int, None]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_generator_class_collections_abc_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import collections.abc

    class Foo(collections.abc.AsyncGenerator[str, int]):
        pass
    """

    expected = """
    import collections.abc

    class Foo(collections.abc.Generator[str, int, None]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterable_annotation(transformer: TreeTransformer) -> None:
    source = """
    from typing import AsyncIterable
    x: AsyncIterable[str, int]
    """

    expected = """
    from typing import AsyncIterable, Iterable
    x: Iterable[str, int]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterable_annotation_collections_abc(
    transformer: TreeTransformer,
) -> None:
    source = """
    from collections.abc import AsyncIterable
    x: AsyncIterable[str, int]
    """

    expected = """
    from collections.abc import AsyncIterable, Iterable
    x: Iterable[str, int]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterable_annotation_typing_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import typing
    x: typing.AsyncIterable[str, int]
    """

    expected = """
    import typing
    x: typing.Iterable[str, int]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterable_annotation_collections_abc_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import collections.abc
    x: collections.abc.AsyncIterable[str, int]
    """

    expected = """
    import collections.abc
    x: collections.abc.Iterable[str, int]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterable_class(transformer: TreeTransformer) -> None:
    source = """
    from typing import AsyncIterable

    class Foo(AsyncIterable[str, int]):
        pass
    """

    expected = """
    from typing import AsyncIterable, Iterable

    class Foo(Iterable[str, int]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterable_class_collections_abc(transformer: TreeTransformer) -> None:
    source = """
    from collections.abc import AsyncIterable

    class Foo(AsyncIterable[str, int]):
        pass
    """

    expected = """
    from collections.abc import AsyncIterable, Iterable

    class Foo(Iterable[str, int]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)  #


def test_async_iterable_class_typing_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import typing

    class Foo(typing.AsyncIterable[str, int]):
        pass
    """

    expected = """
    import typing

    class Foo(typing.Iterable[str, int]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterable_class_collections_abc_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import collections.abc

    class Foo(collections.abc.AsyncIterable[str, int]):
        pass
    """

    expected = """
    import collections.abc

    class Foo(collections.abc.Iterable[str, int]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_typing_import_not_stdlib(transformer: TreeTransformer) -> None:
    source = """
    from foo import typing

    x = typing.AsyncIterable
    """

    expected = """
    from foo import typing

    x = typing.AsyncIterable
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterator_annotation(transformer: TreeTransformer) -> None:
    source = """
    from typing import AsyncIterator
    x: AsyncIterator[str, int]
    """

    expected = """
    from typing import AsyncIterator, Iterator
    x: Iterator[str, int]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterator_annotation_typing_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import typing
    x: typing.AsyncIterator[str, int]
    """

    expected = """
    import typing
    x: typing.Iterator[str, int]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterator_class(transformer: TreeTransformer) -> None:
    source = """
    from typing import AsyncIterator

    class Foo(AsyncIterator[str, int]):
        pass
    """

    expected = """
    from typing import AsyncIterator, Iterator

    class Foo(Iterator[str, int]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_iterator_class_typing_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import typing

    class Foo(typing.AsyncIterator[str, int]):
        pass
    """

    expected = """
    import typing

    class Foo(typing.Iterator[str, int]):
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_exit_stack(transformer: TreeTransformer) -> None:
    source = """
    from contextlib import AsyncExitStack

    async with AsyncExitStack() as some_stack_name:
        await some_stack_name.enter_async_context(foo)
        some_stack_name.push_async_exit(bar)
        some_stack_name.push_async_callback(baz)
        await some_stack_name.aclose()
    """

    expected = """
    from contextlib import AsyncExitStack, ExitStack

    with ExitStack() as some_stack_name:
        some_stack_name.enter_context(foo)
        some_stack_name.push(bar)
        some_stack_name.callback(baz)
        some_stack_name.close()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_async_exit_stack_contextlib_module_import(
    transformer: TreeTransformer,
) -> None:
    source = """
    import contextlib

    async with contextlib.AsyncExitStack() as some_stack_name:
        await some_stack_name.enter_async_context(foo)
        some_stack_name.push_async_exit(bar)
        some_stack_name.push_async_callback(baz)
        await some_stack_name.aclose()
    """

    expected = """
    import contextlib

    with contextlib.ExitStack() as some_stack_name:
        some_stack_name.enter_context(foo)
        some_stack_name.push(bar)
        some_stack_name.callback(baz)
        some_stack_name.close()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_task_group(transformer: TreeTransformer) -> None:
    source = """
    from asyncio import TaskGroup

    async with TaskGroup() as some_name:
        some_name.create_task(some_func())

        await something_else()

        async def nested_function():
            call()
            await async_call()
    """

    expected = """
    from asyncio import TaskGroup
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor() as some_name:
        some_name.submit(some_func)

        something_else()

        def nested_function():
            call()
            async_call()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_task_group_with_args(transformer: TreeTransformer) -> None:
    source = """
    from asyncio import TaskGroup

    async with TaskGroup() as some_name:
        some_name.create_task(some_func(1, foo=bar))
    """

    expected = """
    from asyncio import TaskGroup
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor() as some_name:
        some_name.submit(some_func, 1, foo=bar)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_task_alter_name(transformer: TreeTransformer) -> None:
    source = """
    from asyncio import TaskGroup

    async with TaskGroup() as task_group:
        task_group.create_task(some_func())

    def some_other_thing():
        task_group = None
    """

    expected = """
    from asyncio import TaskGroup
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor() as executor:
        executor.submit(some_func)

    def some_other_thing():
        task_group = None
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_task_group_alias_import(transformer: TreeTransformer) -> None:
    source = """
    from asyncio import TaskGroup as Foobar

    async with Foobar() as task_group:
        task_group.create_task(some_func())
    """

    expected = """
    from asyncio import TaskGroup as Foobar
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor() as executor:
        executor.submit(some_func)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_task_group_module_import(transformer: TreeTransformer) -> None:
    source = """
    import asyncio

    async with asyncio.TaskGroup() as task_group:
        task_group.create_task(some_func())
    """

    expected = """
    import asyncio
    import concurrent.futures

    with concurrent.futures.ThreadPoolExecutor() as executor:
        executor.submit(some_func)
    """

    assert transformer(dedent(source)) == dedent(expected)


@pytest.mark.xfail
def test_asyncio_task_group_reassignment(transformer: TreeTransformer) -> None:
    source = """
    from asyncio import TaskGroup

    task_group = TaskGroup()

    async with task_group:
        task_group.create_task(some_func())
    """

    expected = """
    from concurrent.futures import ThreadPoolExecutor
    from asyncio import TaskGroup

    executor = ThreadPoolExecutor()

    with executor:
        executor.submit(some_func)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_task_missing_as_name_raises(transformer: TreeTransformer) -> None:
    source = """
    from asyncio import TaskGroup

    async with TaskGroup():
        pass
    """

    with pytest.raises(ValueError):
        transformer(dedent(source))


def test_task_group_not_from_asyncio_ignored(transformer: TreeTransformer) -> None:
    source = """
    class TaskGroup:
        pass

    async with TaskGroup():
        pass
    """

    expected = """
    class TaskGroup:
        pass

    with TaskGroup():
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_task_group(transformer: TreeTransformer) -> None:
    source = """
    from anyio import create_task_group

    async with create_task_group() as some_name:
        some_name.start_soon(some_func)
        some_name.start(some_other_func)

        await something_else()

        async def nested_function():
            call()
            await async_call()
    """

    expected = """
    from anyio import create_task_group
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor() as some_name:
        some_name.submit(some_func)
        some_name.submit(some_other_func)

        something_else()

        def nested_function():
            call()
            async_call()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_task_group_with_args(transformer: TreeTransformer) -> None:
    source = """
    from anyio import create_task_group

    async with create_task_group() as task_group:
        task_group.start_soon(some_func, 1, 2)
    """

    expected = """
    from anyio import create_task_group
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor() as executor:
        executor.submit(some_func, 1, 2)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_task_alter_name(transformer: TreeTransformer) -> None:
    source = """
    from anyio import create_task_group

    async with create_task_group() as task_group:
        task_group.start_soon(some_func)

    def some_other_thing():
        task_group = None
    """

    expected = """
    from anyio import create_task_group
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor() as executor:
        executor.submit(some_func)

    def some_other_thing():
        task_group = None
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_task_group_alias_import(transformer: TreeTransformer) -> None:
    source = """
    from anyio import create_task_group as something_else

    async with something_else() as task_group:
        task_group.start_soon(some_func)
    """

    expected = """
    from anyio import create_task_group as something_else
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor() as executor:
        executor.submit(some_func)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_task_group_module_import(transformer: TreeTransformer) -> None:
    source = """
    import anyio

    async with anyio.create_task_group() as task_group:
        task_group.start_soon(some_func)
    """

    expected = """
    import anyio
    import concurrent.futures

    with concurrent.futures.ThreadPoolExecutor() as executor:
        executor.submit(some_func)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_task_missing_as_name_raises(transformer: TreeTransformer) -> None:
    source = """
    from anyio import create_task_group

    async with create_task_group():
        pass
    """

    with pytest.raises(ValueError):
        transformer(dedent(source))


def test_asyncio_sleep(transformer: TreeTransformer) -> None:
    source = """
    from asyncio import sleep
    await sleep(1)
    """

    expected = """
    from asyncio import sleep
    from time import sleep
    sleep(1)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_sleep_module_import(transformer: TreeTransformer) -> None:
    source = """
    import asyncio
    await asyncio.sleep(1)
    """

    expected = """
    import asyncio
    import time
    time.sleep(1)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_sleep_0(transformer: TreeTransformer):
    source = """
    import asyncio

    async def foo():
        await asyncio.sleep(0)
        return True
    """

    expected = """
    import asyncio

    def foo():
        return True
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_sleep(transformer: TreeTransformer) -> None:
    source = """
    from anyio import sleep
    await sleep(1)
    """

    expected = """
    from anyio import sleep
    from time import sleep
    sleep(1)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_sleep_module_import(transformer: TreeTransformer) -> None:
    source = """
    import anyio
    await anyio.sleep(1)
    """

    expected = """
    import anyio
    import time
    time.sleep(1)
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_sleep_0(transformer: TreeTransformer):
    source = """
    import anyio

    async def foo():
        await anyio.sleep(0)
        return True
    """

    expected = """
    import anyio

    def foo():
        return True
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_relative_import(transformer: TreeTransformer) -> None:
    source = """
    from . import foo
    await foo()
    """

    expected = """
    from . import foo
    foo()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_add_new_import_with_module_docstring() -> None:
    transformer = TreeTransformer()

    source = '''
    """This is a module level docstring"""
    import asyncio

    async def foo():
        await asyncio.sleep(1)
    '''

    expected = '''
    """This is a module level docstring"""
    import asyncio
    import time

    def foo():
        time.sleep(1)
    '''

    assert transformer(dedent(source)) == dedent(expected)


def test_add_new_import_with_module_docstring_multiline() -> None:
    transformer = TreeTransformer()

    source = '''
    """This is a module level docstring.
    It has multiple lines
    """
    import asyncio

    async def foo():
        await asyncio.sleep(1)
    '''

    expected = '''
    """This is a module level docstring.
    It has multiple lines
    """
    import asyncio
    import time

    def foo():
        time.sleep(1)
    '''

    assert transformer(dedent(source)) == dedent(expected)


def test_type_checking_import() -> None:
    transformer = TreeTransformer(
        extra_name_replacements={"foo.AsyncThing": "bar.SyncThing"},
    )

    source = """
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from foo import AsyncThing

    async def func() -> AsyncThing:
        ...
    """

    expected = """
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from foo import AsyncThing
        from bar import SyncThing

    def func() -> SyncThing:
        ...
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_reuse_existing_import() -> None:
    transformer = TreeTransformer(
        extra_name_replacements={"foo.AsyncThing": "bar.SyncThing"},
    )

    source = """
    from foo import AsyncThing
    from bar import SyncThing

    async def func() -> AsyncThing:
        ...
    """

    expected = """
    from foo import AsyncThing
    from bar import SyncThing

    def func() -> SyncThing:
        ...
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_add_to_existing_from_import() -> None:
    transformer = TreeTransformer(
        extra_name_replacements={"foo.async_func": "foo.sync_func"},
    )
    source = """
    from foo import bar, async_func

    async def call_func():
        await async_func()
    """

    expected = """
    from foo import bar, async_func, sync_func

    def call_func():
        sync_func()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_unused_name_not_replaced() -> None:
    transformer = TreeTransformer()
    source = """
    from typing import AsyncIterable, Union

    x: AsyncIterable
    """

    expected = """
    from typing import AsyncIterable, Union, Iterable

    x: Iterable
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_unwrap_awaitable_annotation(transformer: TreeTransformer) -> None:
    source = """
    from typing import Awaitable, Callable
    from collections import abc


    async def foo() -> Awaitable[None]:
        pass

    async def bar(fn: Callable[[], Awaitable[int]]) -> None:
        pass

    async def baz() -> abc.Awaitable[None]:
        pass

    aw: Awaitable[list[dict[str, int]]]
    """

    expected = """
    from typing import Awaitable, Callable
    from collections import abc


    def foo() -> None:
        pass

    def bar(fn: Callable[[], int]) -> None:
        pass

    def baz() -> None:
        pass

    aw: list[dict[str, int]]
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_exclude_function() -> None:
    transformer = TreeTransformer(exclude=["foo"])

    source = """
    async def foo():
        pass

    async def bar():
        pass
    """

    expected = """
    async def foo():
        pass

    def bar():
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_exclude_function_nested() -> None:
    transformer = TreeTransformer(exclude=["foo.bar"])

    source = """
    async def foo():
        async def bar():
            pass
    """

    expected = """
    def foo():
        async def bar():
            pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_exclude_class() -> None:
    transformer = TreeTransformer(exclude=["Foo"])

    source = """
    class Foo:
        async def bar() -> None:
            pass

        async def baz() -> None:
            pass

    class Bar:
        async def foo() -> None:
            pass
    """

    expected = """
    class Foo:
        async def bar() -> None:
            pass

        async def baz() -> None:
            pass

    class Bar:
        def foo() -> None:
            pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_exclude_class_method() -> None:
    transformer = TreeTransformer(exclude=["Foo.baz"])

    source = """
    class Foo:
        async def bar() -> None:
            pass

        async def baz() -> None:
            pass
    """

    expected = """
    class Foo:
        def bar() -> None:
            pass

        async def baz() -> None:
            pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_exclude_nested_class() -> None:
    transformer = TreeTransformer(exclude=["Foo.Bar.baz"])

    source = """
    class Foo:
        class Bar:
            async def bar() -> None:
                pass

            async def baz() -> None:
                pass

        async def foo() -> None:
            pass
    """

    expected = """
    class Foo:
        class Bar:
            def bar() -> None:
                pass

            async def baz() -> None:
                pass

        def foo() -> None:
            pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_function_docstring(transformer: TreeTransformer) -> None:
    source = '''
    async def foo():
        """Some text with a ``await this()`` call.
        And here is another line.
        """
        pass
    '''

    expected = '''
    def foo():
        """Some text with a ``this()`` call.
        And here is another line.
        """
        pass
    '''

    assert transformer(dedent(source)) == dedent(expected)


def test_class_docstring(transformer: TreeTransformer) -> None:
    source = '''
    class Foo:
        """Some text with a ``await this()`` call.
        And here is another line.
        """
    '''

    expected = '''
    class Foo:
        """Some text with a ``this()`` call.
        And here is another line.
        """
    '''

    assert transformer(dedent(source)) == dedent(expected)


def test_module_docstring(transformer: TreeTransformer) -> None:
    source = '''
    """Some text with a ``await this()`` call.
    And here is another line.
    """
    '''

    expected = '''
    """Some text with a ``this()`` call.
    And here is another line.
    """
    '''

    assert transformer(dedent(source)) == dedent(expected)


def test_wrapped_await_expression_preserve_inline_comment(
    transformer: TreeTransformer,
) -> None:
    source = """
    foo = (  # type:ignore[no-any-return]  # pragma: no cover
        await something()
    )
    """

    expected = """
    foo = (  # type:ignore[no-any-return]  # pragma: no cover
        something()
    )
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_function_preserve_inline_comment(transformer: TreeTransformer) -> None:
    source = """
    async def foo(  # preserve this
    ) -> None:  # and this
        pass
    """

    expected = """
    def foo(  # preserve this
    ) -> None:  # and this
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_honour_type_checking_import() -> None:
    transformer = TreeTransformer(
        extra_name_replacements={
            "module_a.AsyncThing": "module_b.SyncThing",
            "async_module.Something": "sync_module.SomethingElse",
        },
    )

    source = """
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import async_module
        from module_a import AsyncThing

    def bar(param: async_module.Something) -> AsyncThing:
        pass
    """

    expected = """
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import async_module
        from module_a import AsyncThing
        from module_b import SyncThing
        import sync_module

    def bar(param: sync_module.SomethingElse) -> SyncThing:
        pass
    """
    assert transformer(dedent(source)) == dedent(expected)


def test_honour_type_checking_import_submodule() -> None:
    transformer = TreeTransformer(
        extra_name_replacements={
            "module_a.submodule_a.AsyncThing": "module_b.submodule_b.SyncThing",
            "module_c.sub_c.ThingC": "module_d.sub_d.ThingD",
        },
    )

    source = """
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from module_a.submodule_a import AsyncThing
        import module_c.sub_c

    def bar(param: module_c.sub_c.ThingC) -> AsyncThing:
        pass
    """

    expected = """
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from module_a.submodule_a import AsyncThing
        import module_c.sub_c
        from module_b.submodule_b import SyncThing
        import module_d.sub_d

    def bar(param: module_d.sub_d.ThingD) -> SyncThing:
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_honour_type_checking_import_multiple_names_from_same_module() -> None:
    transformer = TreeTransformer(
        extra_name_replacements={
            "foo.ThingOne": "bar.ThingThree",
            "foo.ThingTwo": "bar.ThingFour",
        },
    )

    source = """
    from typing import TYPE_CHECKING
    from foo import ThingOne

    if TYPE_CHECKING:
        from foo import ThingTwo

    def baz(param: ThingOne) -> ThingTwo:
        pass
    """

    expected = """
    from typing import TYPE_CHECKING
    from foo import ThingOne
    from bar import ThingThree

    if TYPE_CHECKING:
        from foo import ThingTwo
        from bar import ThingFour

    def baz(param: ThingThree) -> ThingFour:
        pass
    """
    result = transformer(dedent(source))
    assert result == dedent(expected)


def test_disable_infer_type_checking_imports() -> None:
    transformer = TreeTransformer(
        extra_name_replacements={
            "module_a.AsyncThing": "module_b.SyncThing",
        },
        infer_type_checking_imports=False,
    )

    source = """
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from module_a import AsyncThing

    def bar() -> AsyncThing:
        pass
    """

    expected = """
    from typing import TYPE_CHECKING
    from module_b import SyncThing

    if TYPE_CHECKING:
        from module_a import AsyncThing

    def bar() -> SyncThing:
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_preserve_import_formatting(transformer: TreeTransformer) -> None:
    source = """
    from typing import (
        AsyncGenerator,
        Generator,
    )
    """

    assert transformer(dedent(source)) == dedent(source)


def test_ruff_fix(tmp_path, monkeypatch) -> None:
    config_file = tmp_path / "ruff.toml"
    config_file.write_text(tomli_w.dumps({"lint": {"select": ["I001"]}}))
    monkeypatch.chdir(tmp_path)
    transformer = TreeTransformer(ruff_fix=True)

    source = """
    import time, asyncio
    """

    expected = """
    import asyncio
    import time
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_ruff_fix_ruff_error_raises(tmp_path, monkeypatch) -> None:
    config_file = tmp_path / "ruff.toml"
    config_file.write_text(tomli_w.dumps({"lint": {"select": ["FOO017"]}}))
    monkeypatch.chdir(tmp_path)
    transformer = TreeTransformer(ruff_fix=True)

    source = """
    import asyncio
    """

    with pytest.raises(ChildProcessError, match="Error calling ruff"):
        transformer(dedent(source))


def test_ruff_fix_pass_file_name(tmp_path, monkeypatch) -> None:
    config_file = tmp_path / "ruff.toml"
    config_file.write_text(
        tomli_w.dumps(
            {
                "lint": {
                    "per-file-ignores": {"some_file.py": ["I001"]},
                    "select": ["I001"],
                }
            }
        )
    )
    monkeypatch.chdir(tmp_path)
    transformer = TreeTransformer(ruff_fix=True, file_name="some_file.py")

    source = """
    import time, asyncio
    """

    expected = """
    import time, asyncio
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_ruff_format(tmp_path, monkeypatch) -> None:
    config_file = tmp_path / "ruff.toml"
    config_file.write_text(tomli_w.dumps({"format": {"quote-style": "single"}}))
    monkeypatch.chdir(tmp_path)
    transformer = TreeTransformer(ruff_format=True)

    source = """mode = "format"
    """

    expected = """mode = 'format'
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_ruff_format_ruff_error_raises(tmp_path, monkeypatch):
    config_file = tmp_path / "ruff.toml"
    config_file.write_text(tomli_w.dumps({"format": {"quote-style": "quadruple"}}))
    monkeypatch.chdir(tmp_path)
    transformer = TreeTransformer(ruff_format=True)

    source = """
    mode = "format"
    """

    with pytest.raises(ChildProcessError, match="Error calling ruff"):
        transformer(dedent(source))


def test_ruff_fix_and_format(tmp_path, monkeypatch) -> None:
    config_file = tmp_path / "ruff.toml"
    config_file.write_text(
        tomli_w.dumps(
            {"format": {"quote-style": "single"}, "lint": {"select": ["I001"]}}
        )
    )
    monkeypatch.chdir(tmp_path)
    transformer = TreeTransformer(ruff_fix=True, ruff_format=True)

    source = """
    import time, asyncio
    mode = "format"
    """

    expected = """
    import asyncio
    import time

    mode = 'format'
    """

    assert transformer(dedent(source)) == dedent(expected).lstrip()


def test_async_comprehension(transformer: TreeTransformer) -> None:
    source = """
    [x async for x in foo()]
    {x async for x in foo()}
    (x async for x in foo())
    {x: 1 async for x in foo()}
    """

    expected = """
    [x for x in foo()]
    {x for x in foo()}
    (x for x in foo())
    {x: 1 for x in foo()}
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_replace_relative_imported_names() -> None:
    transformer = TreeTransformer(
        extra_name_replacements={".foo.AsyncThing": ".bar.SyncThing"}
    )
    source = """
    from .foo import AsyncThing

    async def func() -> AsyncThing:
        ...
    """

    expected = """
    from .foo import AsyncThing
    from .bar import SyncThing

    def func() -> SyncThing:
        ...
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_module_dunder_all_import():
    transformer = TreeTransformer(
        extra_name_replacements={
            ".foo.AsyncThing": ".foo.SyncThing",
            "bar.Something": "baz.SomethingElse",
        }
    )

    source = """
    from .foo import AsyncThing
    from bar import Something

    __all__ = ("AsyncThing", "Something")
    """

    expected = """
    from .foo import AsyncThing, SyncThing
    from bar import Something
    from baz import SomethingElse

    __all__ = ("SyncThing", "SomethingElse")
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_path(transformer):
    source = """
    from anyio import Path

    foo = await Path().read_text()
    """

    expected = """
    from anyio import Path
    from pathlib import Path

    foo = Path().read_text()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_path_module_import(transformer):
    source = """
    import anyio

    foo = await anyio.Path().read_text()
    """

    expected = """
    import anyio
    import pathlib

    foo = pathlib.Path().read_text()
    """

    assert transformer(dedent(source)) == dedent(expected)


@pytest.mark.xfail(reason="not implemented")
def test_transform_imports_in_function(transformer):
    source = """
    def foo():
        from contextlib import asynccontextmanager
    """

    expected = """
    def foo():
        from contextlib import asynccontextmanager
        from contextlib import contextmanager
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_semaphore(transformer):
    source = """
    import asyncio

    sem = asyncio.Semaphore(2)

    await sem.acquire()
    await sem.release()

    async with sem:
        pass
    """

    expected = """
    import asyncio
    import threading

    sem = threading.Semaphore(2)

    sem.acquire()
    sem.release()

    with sem:
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_semaphore(transformer):
    source = """
    import anyio

    sem = anyio.Semaphore(2)

    await sem.acquire()
    await sem.release()

    async with sem:
        pass
    """

    expected = """
    import anyio
    import threading

    sem = threading.Semaphore(2)

    sem.acquire()
    sem.release()

    with sem:
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_lock(transformer):
    source = """
    import asyncio

    lock = asyncio.Lock()

    async with lock:
        pass
    """

    expected = """
    import asyncio
    import threading

    lock = threading.Lock()

    with lock:
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_lock(transformer):
    source = """
    import anyio

    lock = anyio.Lock()

    async with lock:
        pass
    """

    expected = """
    import anyio
    import threading

    lock = threading.Lock()

    with lock:
        pass
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_event(transformer):
    source = """
    import asyncio

    event = asyncio.Event()

    event.set()
    await event.wait()
    assert event.is_set()
    event.clear()
    """

    expected = """
    import asyncio
    import threading

    event = threading.Event()

    event.set()
    event.wait()
    assert event.is_set()
    event.clear()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_anyio_event(transformer):
    source = """
    import anyio

    event = anyio.Event()

    event.set()
    await event.wait()
    assert event.is_set()
    """

    expected = """
    import anyio
    import threading

    event = threading.Event()

    event.set()
    event.wait()
    assert event.is_set()
    """

    assert transformer(dedent(source)) == dedent(expected)


def test_asyncio_barrier(transformer):
    source = """
    import asyncio

    barrier = asyncio.Barrier(2)

    await barrier.wait()
    await barrier.reset()
    await barrier.abort()

    barrier.parties
    barrier.n_waiting
    barrier.broken
    """

    expected = """
    import asyncio
    import threading

    barrier = threading.Barrier(2)

    barrier.wait()
    barrier.reset()
    barrier.abort()

    barrier.parties
    barrier.n_waiting
    barrier.broken
    """

    assert transformer(dedent(source)) == dedent(expected)



import secrets
import shutil
import textwrap
from pathlib import Path
from unittest.mock import MagicMock

import pytest
import tomli_w
from click.testing import CliRunner
from pytest import MonkeyPatch
from pytest_mock import MockerFixture

from unasyncd.cli import main
from unasyncd.main import Env


@pytest.fixture(scope="session", autouse=True)
def clear_cache():
    shutil.rmtree(".unasyncd_cache", ignore_errors=True)
    yield
    shutil.rmtree(".unasyncd_cache", ignore_errors=True)


TEST_CONTENT = textwrap.dedent(
    """
async def foo() -> None:
    pass
"""
)

TEST_TRANSFORMED_CONTENT = textwrap.dedent(
    """
def foo() -> None:
    pass
"""
)


@pytest.fixture()
def mock_cache(mocker: MockerFixture):
    return mocker.patch("unasyncd.main.Cache")


@pytest.fixture()
def source_file(tmp_path) -> Path:
    path = (tmp_path / secrets.token_hex()).with_suffix(".py")
    path.write_text(TEST_CONTENT)
    return path


@pytest.fixture()
def target_file(tmp_path) -> Path:
    return tmp_path.joinpath(secrets.token_hex()).with_suffix(".py")


@pytest.fixture()
def mock_transform(mocker: MockerFixture) -> MagicMock:
    return mocker.spy(Env, "transform")


@pytest.fixture
def runner() -> CliRunner:
    return CliRunner()


@pytest.fixture()
def mock_unasync_files(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("unasyncd.cli.unasync_files")


@pytest.fixture
def mock_load_config(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("unasyncd.cli.load_config")


@pytest.fixture()
def mock_string_transformer(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("unasyncd.main.StringTransformer").return_value


@pytest.fixture()
def mock_tree_transformer(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("unasyncd.tree_transformer.TreeTransformer").return_value


@pytest.mark.parametrize("config_file", ["pyproject.toml", ".unasyncd.toml"])
@pytest.mark.usefixtures("mock_unasync_files")
def test_default_config_file(
    runner: CliRunner,
    tmp_path: Path,
    mock_load_config: MagicMock,
    config_file: str,
    monkeypatch: MonkeyPatch,
):
    monkeypatch.chdir(tmp_path)
    config_path = tmp_path / config_file
    config_path.touch()

    result = runner.invoke(main)

    assert not result.exception
    mock_load_config.assert_called_once()
    assert mock_load_config.call_args_list[0].kwargs["path"].absolute() == config_path


@pytest.mark.usefixtures("mock_unasync_files")
def test_set_config_file(
    runner: CliRunner, tmp_path: Path, mock_load_config: MagicMock
):
    config_path = tmp_path / "pyproject.toml"

    result = runner.invoke(main, ["-c", str(config_path)])

    assert not result.exception
    mock_load_config.assert_called_once()
    assert mock_load_config.call_args_list[0].kwargs["path"] == config_path


@pytest.mark.parametrize(
    "flag_value, config_value",
    [
        (True, True),
        (False, False),
        (True, False),
        (False, True),
    ],
)
def test_feature_flags(
    runner: CliRunner,
    mock_unasync_files: MagicMock,
    source_file: Path,
    target_file: Path,
    tmp_path: Path,
    monkeypatch: MonkeyPatch,
    flag_value: bool,
    config_value: bool,
):
    monkeypatch.chdir(tmp_path)
    arguments = [f"{source_file}:{target_file}"]

    arguments.extend(
        [
            f"--{'no-' if not flag_value else ''}{flag}"
            for flag in [
                "transform-docstrings",
                "add-editors-note",
                "infer-type-checking-imports",
                "force",
                "cache",
                "ruff-fix",
            ]
        ]
    )

    arguments.append("--check" if flag_value else "--write")

    result = runner.invoke(main, arguments)

    assert not result.exception
    assert result.exit_code == 0
    config = mock_unasync_files.call_args_list[0].kwargs["config"]

    assert config.transform_docstrings is flag_value
    assert config.add_editors_note is flag_value
    assert config.force_regen is flag_value
    assert config.infer_type_checking_imports is flag_value
    assert config.cache is flag_value
    assert config.check_only is flag_value


def test_pass_files(
    runner: CliRunner,
    mock_unasync_files: MagicMock,
    tmp_path: Path,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    tmp_path.joinpath("foo.py").touch()
    tmp_path.joinpath("bar.py").touch()

    result = runner.invoke(main, ["foo.py:foo_sync.py", "bar.py:bar_sync.py"])

    assert not result.exception
    config = mock_unasync_files.call_args_list[0].kwargs["config"]

    assert config.files == {"foo.py": "foo_sync.py", "bar.py": "bar_sync.py"}


@pytest.mark.usefixtures("mock_unasync_files")
@pytest.mark.parametrize("files_transformed", [True, False])
def test_return_code(
    runner: CliRunner,
    mocker: MockerFixture,
    files_transformed: bool,
    source_file: Path,
    target_file: Path,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.chdir(source_file.parent)
    mock_run = mocker.patch("unasyncd.cli._run")
    mock_run.return_value = files_transformed

    result = runner.invoke(main, f"{source_file}:{target_file}")
    assert result.exit_code == int(files_transformed)


def test_transform(runner: CliRunner, source_file: Path, target_file: Path) -> None:
    result = runner.invoke(main, [f"{source_file}:{target_file}"])

    assert result.exit_code == 1
    assert target_file.read_text() == TEST_TRANSFORMED_CONTENT


def test_transform_source_and_target_unchanged(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [
        f"{source_file}:{target_file}",
    ]
    result_1 = runner.invoke(main, args)
    result_2 = runner.invoke(main, args)

    assert result_1.exit_code == 1
    assert result_2.exit_code == 0
    assert mock_transform.call_count == 1
    assert target_file.read_text() == TEST_TRANSFORMED_CONTENT


def test_transform_regen_config_changed(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [f"{source_file}:{target_file}"]
    source_file.write_text(
        textwrap.dedent(
            """
    async def foo():
        '''this is a docstring with await bar()'''
    """
        )
    )

    expected_content = textwrap.dedent(
        """
    def foo():
        '''this is a docstring with bar()'''
    """
    )

    target_file.write_text(expected_content)

    result_1 = runner.invoke(main, args)
    result_2 = runner.invoke(main, [*args, "--transform-docstrings"])

    assert result_1.exit_code == 1
    assert result_2.exit_code == 1
    assert mock_transform.call_count == 2
    assert target_file.read_text() == expected_content


def test_transform_source_and_target_unchanged_force_regen(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [f"{source_file}:{target_file}"]
    result_1 = runner.invoke(main, args)
    assert target_file.read_text() == TEST_TRANSFORMED_CONTENT
    assert result_1.exit_code == 1

    result_2 = runner.invoke(main, [*args, "--force"])

    assert result_2.exit_code == 0
    assert mock_transform.call_count == 2
    assert target_file.read_text() == TEST_TRANSFORMED_CONTENT


def test_transform_cached(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [f"{source_file}:{target_file}"]
    result_1 = runner.invoke(main, args)
    target_file.unlink()

    result_2 = runner.invoke(main, args)

    assert result_1.exit_code == 1
    assert result_2.exit_code == 1
    assert mock_transform.call_count == 1
    assert target_file.read_text() == TEST_TRANSFORMED_CONTENT


def test_transform_cached_no_cache(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [f"{source_file}:{target_file}", "--no-cache"]

    result_1 = runner.invoke(main, args)

    target_file.unlink()

    result_2 = runner.invoke(main, args)

    assert result_1.exit_code == 1
    assert result_2.exit_code == 1
    assert mock_transform.call_count == 2
    assert target_file.read_text() == TEST_TRANSFORMED_CONTENT


def test_transform_cached_force_regen(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [f"{source_file}:{target_file}", "--force"]

    result_1 = runner.invoke(main, args)

    target_file.unlink()

    result_2 = runner.invoke(main, args)

    assert result_1.exit_code == 1
    assert result_2.exit_code == 1
    assert mock_transform.call_count == 2
    assert target_file.read_text() == TEST_TRANSFORMED_CONTENT


def test_transform_source_changed_ast_stable(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [
        f"{source_file}:{target_file}",
    ]
    result_1 = runner.invoke(main, args)

    source_file.write_text(
        textwrap.dedent(
            """
    async def   foo() -> None :
        pass
    """
        )
    )

    result_2 = runner.invoke(main, args)

    assert result_1.exit_code == 1
    assert result_2.exit_code == 0
    assert mock_transform.call_count == 1
    assert target_file.read_text() == TEST_TRANSFORMED_CONTENT


def test_transform_target_changed_ast_stable(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [
        f"{source_file}:{target_file}",
    ]
    result_1 = runner.invoke(main, args)

    target_content = textwrap.dedent(
        """
    def   foo() -> None :
        pass
    """
    )
    target_file.write_text(target_content)

    result_2 = runner.invoke(main, args)

    assert result_1.exit_code == 1
    assert result_2.exit_code == 0
    assert mock_transform.call_count == 1
    assert target_file.read_text() == target_content


def test_transform_source_and_target_changed_ast_stable(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [
        f"{source_file}:{target_file}",
    ]
    result_1 = runner.invoke(main, args)

    source_file.write_text(
        textwrap.dedent(
            """
    async def   foo() -> None :
        pass
    """
        )
    )

    target_content = textwrap.dedent(
        """
    def   foo() -> None :
        pass
    """
    )
    target_file.write_text(target_content)

    result_2 = runner.invoke(main, args)

    assert result_1.exit_code == 1
    assert result_2.exit_code == 0
    assert mock_transform.call_count == 1
    assert target_file.read_text() == target_content


def test_transform_source_and_target_changed_no_transformation(
    runner: CliRunner, source_file: Path, target_file: Path, mock_transform: MagicMock
) -> None:
    args = [f"{source_file}:{target_file}"]
    result_1 = runner.invoke(main, args, catch_exceptions=False)

    target_file.write_text(TEST_TRANSFORMED_CONTENT + "\n\nimport foo")
    source_file.write_text(TEST_CONTENT + "\n\nimport foo")

    result_2 = runner.invoke(main, args, catch_exceptions=False)

    assert result_1.exit_code == 1
    assert result_2.exit_code == 0
    assert mock_transform.call_count == 2


def test_transform_add_editors_note(
    runner: CliRunner, source_file: Path, target_file: Path
) -> None:
    result = runner.invoke(
        main,
        [f"{source_file}:{target_file}", "--add-editors-note"],
        catch_exceptions=False,
    )

    assert result.exit_code == 1
    expected_content = (
        f"# Do not edit this file directly. It has been autogenerated "
        f"from\n# {source_file}\n{TEST_TRANSFORMED_CONTENT}"
    )
    assert target_file.read_text() == expected_content


def test_ruff_fix_pass_file_name(tmp_path, monkeypatch, runner: CliRunner) -> None:
    source_file = tmp_path / "some_file.py"
    target_file = tmp_path / "some_other_file.py"
    config_file = tmp_path / "ruff.toml"
    config_file.write_text(
        tomli_w.dumps(
            {
                "lint": {
                    "per-file-ignores": {"some_file.py": ["I001"]},
                    "select": ["I001"],
                }
            }
        )
    )
    monkeypatch.chdir(tmp_path)

    source = """
    import time, asyncio
    """
    source_file.write_text(textwrap.dedent(source))

    expected = """
    import time, asyncio
    """

    runner.invoke(
        main,
        [f"{source_file}:{target_file}", "--ruff-fix"],
        catch_exceptions=False,
    )
    assert target_file.read_text() == textwrap.dedent(expected)



from pathlib import Path

import msgspec
import pytest

from unasyncd.config import Config, load_config


@pytest.fixture
def config_from_file(tmp_path: Path, monkeypatch):
    path = tmp_path / "pyproject.toml"

    def _create_file(**kwargs) -> Config:
        path.write_bytes(msgspec.toml.encode({"tool": {"unasyncd": kwargs}}))
        monkeypatch.chdir(tmp_path)
        return load_config(path)

    return _create_file


def test_default_config(config_from_file):
    config = config_from_file()

    assert config.add_editors_note is False
    assert config.transform_docstrings is False
    assert config.cache is True
    assert config.check_only is False
    assert config.force_regen is False
    assert config.files == {}
    assert config.exclude == {}
    assert config.extra_replacements == {}
    assert config.infer_type_checking_imports is True
    assert config.ruff_fix is False


def test_config_override(config_from_file, tmp_path):
    tmp_path.joinpath("foo.py").touch()

    config = config_from_file(
        add_editors_note=True,
        transform_docstrings=True,
        cache=False,
        check_only=True,
        force_regen=True,
        files={"foo.py": "bar.py"},
        infer_type_checking_imports=False,
        ruff_fix=True,
    )

    assert config.add_editors_note is True
    assert config.transform_docstrings is True
    assert config.cache is False
    assert config.check_only is True
    assert config.force_regen is True
    assert config.files == {"foo.py": "bar.py"}
    assert config.infer_type_checking_imports is False
    assert config.ruff_fix is True


def test_file_directories(config_from_file, tmp_path):
    source_dir = tmp_path / "foo"
    sub_dir = source_dir / "bar"
    file_1 = source_dir / "one.py"
    file_2 = source_dir / "two.py"
    file_3 = sub_dir / "three.py"

    for file in [file_1, file_2, file_3]:
        file.parent.mkdir(parents=True, exist_ok=True)
        file.touch()

    config = config_from_file(files={"foo": "foo_sync"})

    assert config.files == {
        "foo/bar/three.py": "foo_sync/bar/three.py",
        "foo/one.py": "foo_sync/one.py",
        "foo/two.py": "foo_sync/two.py",
    }


def test_excludes(config_from_file, tmp_path):
    tmp_path.joinpath("this.py").touch()

    config = config_from_file(
        files={"this.py": "that.py"},
        exclude=["Some.thing"],
        per_file_exclude={
            "foo.py": ["Something.else"],
            "bar.py": ["this", "And.that"],
        },
    )
    assert config.exclude == {
        "this.py": ["Some.thing"],
        "foo.py": ["Some.thing", "Something.else"],
        "bar.py": ["Some.thing", "this", "And.that"],
    }


def test_add_replacements(config_from_file, tmp_path):
    tmp_path.joinpath("this.py").touch()
    tmp_path.joinpath("that.py").touch()

    config = config_from_file(
        files={"this.py": "that.py"},
        add_replacements={"async_foo": "sync_foo"},
        per_file_add_replacements={
            "foo.py": {"this": "that"},
            "bar/baz.py": {"something": "else", "_and": "more_"},
        },
    )

    assert config.extra_replacements == {
        "this.py": {"async_foo": "sync_foo"},
        "foo.py": {"async_foo": "sync_foo", "this": "that"},
        "bar/baz.py": {"async_foo": "sync_foo", "something": "else", "_and": "more_"},
    }


def test_config_key(config_from_file):
    config = config_from_file()
    assert config.key() == config.key()

    old_key = config.key()
    config.force_regen = True
    assert config.key() != old_key


def test_config_key_changes_with_version(config_from_file, monkeypatch):
    config = config_from_file()
    old_key = config.key()

    import unasyncd.config

    monkeypatch.setattr(unasyncd.config, "VERSION", "99.0.0")

    assert old_key != config.key()



from __future__ import annotations

import asyncio
import time
from contextlib import nullcontext
from pathlib import Path

import rich
import rich_click as click

from .config import Config, load_config
from .main import unasync_files


async def _run(*, config: Config, check_only: bool, verbose: bool) -> bool:
    console = rich.get_console()
    verbose_console = rich.console.Console()
    verbose_console.quiet = not verbose

    start = time.perf_counter()
    files_changed = 0
    files_unchanged = 0

    status = console.status("Processing") if not verbose else nullcontext()
    with status:
        async for result in unasync_files(config=config):
            if not result.transformed:
                files_unchanged += 1
                continue

            files_changed += 1

            verbose_console.print(
                f"Transformed [yellow]{result.source}[/] > [green]{result.target}[/]"
            )

    console.print(f"Finished in {round(time.perf_counter() - start, 2)} seconds")
    console.print(
        f"[green]{files_changed}[/] file{'s' if files_changed != 1 else ''} "
        f"{'would be ' if check_only else ''}transformed, "
        f"[blue]{files_unchanged}[/] {'would be ' if check_only else ''}"
        "left unchanged"
    )

    return files_changed > 0


@click.command()
@click.option(
    "--cache/--no-cache",
    is_flag=True,
    help="Cache transformation results",
    default=None,
)
@click.option(
    "--transform-docstrings/--no-transform-docstrings",
    is_flag=True,
    default=None,
    help="Transform module, class, method and function docstrings",
)
@click.option(
    "--infer-type-checking-imports/--no-infer-type-checking-imports",
    is_flag=True,
    default=None,
    help="Infer if new imports should be added to an 'if TYPE_CHECKING' block",
)
@click.option(
    "--add-editors-note/--no-add-editors-note",
    is_flag=True,
    default=None,
    help="Add an editors note to the generated files",
)
@click.option(
    "--ruff-fix/--no-ruff-fix",
    is_flag=True,
    default=None,
    help="Run 'ruff --fix' on the transformed output before writing it back",
)
@click.option(
    "--ruff-format/--no-ruff-format",
    is_flag=True,
    default=None,
    help="Run 'ruff format' on the transformed output before writing it back",
)
@click.option(
    "--check/--write",
    "check_only",
    is_flag=True,
    default=None,
    help="Write changes back",
)
@click.option(
    "-c",
    "--config",
    "config_file",
    help="Configuration file. If not given defaults to pyproject.toml",
    type=click.Path(dir_okay=False, path_type=Path),
)
@click.option(
    "-f/-F",
    "--force/--no-force",
    "force_regen",
    help="Force regeneration regardless of changes",
    is_flag=True,
    default=None,
)
@click.option(
    "-v", "--verbose", is_flag=True, help="Increase verbosity of console output"
)
@click.option("-q", "--quiet", is_flag=True, help="Suppress all console output")
@click.argument("files", nargs=-1)
def main(
    files: tuple[str, ...],
    cache: bool | None,
    transform_docstrings: bool | None,
    infer_type_checking_imports: bool | None,
    ruff_fix: bool | None,
    ruff_format: bool | None,
    check_only: bool,
    add_editors_note: bool | None,
    config_file: Path | None,
    force_regen: bool | None,
    verbose: bool,
    quiet: bool,
) -> None:
    console = rich.get_console()
    console.quiet = quiet

    if config_file is None:
        for name in ["pyproject.toml", ".unasyncd.toml"]:
            path = Path(name)
            if path.exists():
                config_file = path

    if config_file:
        console.print(f"[cyan]Config file: [yellow][b]{config_file}")

    config = load_config(
        path=config_file,
        cache=cache,
        transform_docstrings=transform_docstrings,
        add_editors_note=add_editors_note,
        files=dict(f.split(":", 1) for f in files) if files else None,
        check_only=check_only,
        force_regen=force_regen,
        infer_type_checking_imports=infer_type_checking_imports,
        ruff_fix=ruff_fix,
        ruff_format=ruff_format,
    )

    if not config.files:
        console.print("[red]No files selected. Quitting")
    else:
        quit(
            int(
                asyncio.run(_run(config=config, check_only=check_only, verbose=verbose))
            )
        )
